# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='FollowRecord',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField(db_index=True)),
                ('content_type', models.ForeignKey(related_name='followed_record_contenttype', to='contenttypes.ContentType')),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



"""
Test cases for the follower module
"""
from django.test import TestCase
from followit.compat import get_user_model

class FollowerTests(TestCase):
    def setUp(self):
        User = get_user_model()
        self.u1 = User.objects.create_user('user1', 'user1@example.com')
        self.u2 = User.objects.create_user('user2', 'user2@example.com')
        self.u3 = User.objects.create_user('user3', 'user3@example.com')

    def test_multiple_follow(self):
        
        self.u1.follow_user(self.u2)
        self.u1.follow_user(self.u3)
        self.u2.follow_user(self.u1)

        self.assertEquals(
            set(self.u1.get_followers()),
            set([self.u2])
        )

        self.assertEquals(
            set(self.u2.get_followers()),
            set([self.u1])
        )

        self.assertEquals(
            set(self.u1.get_followed_users()),
            set([self.u2, self.u3])
        )

    def test_unfollow(self):
        self.u1.follow_user(self.u2)
        self.u1.unfollow_user(self.u2)
        self.assertEquals(self.u1.get_followed_users().count(), 0)

    def test_is_following(self):
        self.u2.follow_user(self.u1)
        self.assertTrue(self.u2.is_following(self.u1))
        self.assertFalse(self.u1.is_following(self.u2))



# Src: https://github.com/toastdriven/django-tastypie/blob/master/tastypie/compat.py
# Note: Appending original license in file itself. License files are hard :(.
#
# Copyright (c) 2010, Daniel Lindsley
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the tastypie nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL tastypie BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
import django

__all__ = ['get_user_model',]

# Django 1.5+ compatibility
def get_user_model():
    if django.VERSION >= (1, 5):
        try:
            from django.contrib import auth
            return auth.get_user_model()
        except ImproperlyConfigured:
            # The the users model might not be read yet.
            # This can happen is when setting up the create_api_key signal, in your
            # custom user module.
            return None
    else:
        from django.contrib.auth.models import User
        return User

USER_MODEL_CLASS_NAME = getattr(settings, 'AUTH_USER_CLASS', 'auth.User')



import django
import os
import sys

RUNTESTS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..', '..')
sys.path.insert(0, os.path.dirname(RUNTESTS_DIR))

os.environ['DJANGO_SETTINGS_MODULE'] = 'followit.tests.settings'
try:
    django.setup()
except AttributeError:  # django < 1.7
    pass

from followit.tests import settings
from django.test.utils import get_runner

TestRunner = get_runner(settings)
test_runner = TestRunner(interactive=False)
failures = test_runner.run_tests(['tests.FollowerTests'])

if failures:
    sys.exit(failures)




"""A Django App allowing :class:`~django.contrib.auth.models.User` to 
follow instances of any other django models, including other users

To use this module:
* add "followit" to the ``INSTALLED_APPS`` in your ``settings.py``
* in your app's ``models.py`` add:

    import followit
    followit.register(Thing)

* run ``python manage.py syncdb``
* then anywhere in your code you can do the following:

    user.follow(some_thing_instance)
    user.unfollow(some_thing_instance)

    user.get_followed_things() #note that "things" is from the name of class Thing
    some_thing.get_followers()

Copyright 2011 Evgeny Fadeev evgeny.fadeev@gmail.com
The source code is available under BSD license.
"""
import django
from followit import utils
from followit.compat import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured

if django.VERSION < (1, 7) or django.VERSION >= (1, 9):
    msg = """This version of django-followit supports django 1.7 and 1.8.
For earlier versions of django django-followit 0.0.9 can be used"""
    raise ImproperlyConfigured(msg)

REGISTRY = {}


def get_model_name(model):
    try:
        return model._meta.module_name
    except AttributeError:
        return model._meta.model_name


def get_follow_records(user, obj):
    from followit.models import FollowRecord
    ct = ContentType.objects.get_for_model(obj)
    return FollowRecord.objects.filter(
                                content_type=ct,
                                object_id=obj.pk,
                                user=user
                            )


def get_object_followers(obj):
    """returns query set of users following the object"""
    from followit.models import FollowRecord
    ct = ContentType.objects.get_for_model(obj)
    fr_set = FollowRecord.objects.filter(content_type=ct, object_id=obj.pk)
    uids = fr_set.values_list('user', flat=True)
    User = get_user_model()
    return User.objects.filter(pk__in=uids)


def make_followed_objects_getter(model):
    """returns query set of objects of a class ``model``
    that are followed by a user"""

    #something like followX_set__user
    def followed_objects_getter(user):
        from followit.models import FollowRecord
        ct = ContentType.objects.get_for_model(model)
        fr_set = FollowRecord.objects.filter(
                                content_type=ct,
                                user=user
                            )
        obj_id_set = fr_set.values_list('object_id', flat=True)
        return model.objects.filter(pk__in=obj_id_set)


    return followed_objects_getter

def test_follow_method(user, obj):
    """True if object ``obj`` is followed by the user,
    false otherwise, no error checking on whether the model
    has or has not been registered with the ``followit`` app
    """
    fr_set = get_follow_records(user, obj)
    return fr_set.exists()


def make_follow_method(model):
    """returns a method that adds a FollowX record
    for an object
    """
    def follow_method(user, obj):
        """returns ``True`` if follow operation created a new record"""
        from followit.models import FollowRecord
        ct = ContentType.objects.get_for_model(obj)
        fr, created = FollowRecord.objects.get_or_create(
                                                    content_type=ct,
                                                    object_id=obj.pk,
                                                    user=user
                                                )
        return created
    return follow_method


def make_unfollow_method(model):
    """returns a method that allows to unfollow an item
    """
    def unfollow_method(user, obj):
        """attempts to find an item and delete it, no
        exstence checking
        """
        fr_set = get_follow_records(user, obj)
        fr_set.delete()
    return unfollow_method


def register(model):
    """returns model class that connects
    User with the followed object

    ``model`` - is the model class to follow

    The ``model`` class gets new method - ``get_followers``
    and the User class - a method - ``get_followed_Xs``, where
    the ``X`` is the name of the model
    and ``is_following(something)``
    
    Note, that proper pluralization of the model name is not supported,
    just "s" is added
    """
    from followit import models as followit_models
    from django.db import models as django_models
    from django.db.models.fields.related import ForeignKey

    User = get_user_model()

    model_name = get_model_name(model)
    if model in REGISTRY:
        return
    REGISTRY[model_name] = model

    #1) patch ``model`` with method ``get_followers()``
    model.add_to_class('get_followers', get_object_followers)

    #2) patch ``User`` with method ``get_followed_Xs``
    method_name = 'get_followed_' + model_name + 's'
    getter_method = make_followed_objects_getter(model)
    User.add_to_class(method_name, getter_method)

    #3) patch ``User with method ``is_following()``
    if not hasattr(User, 'is_following'):
        User.add_to_class('is_following', test_follow_method)

    #4) patch ``User`` with method ``follow_X``
    follow_method = make_follow_method(model)
    User.add_to_class('follow_' + model_name, follow_method)

    #5) patch ``User`` with method ``unfollow_X``
    unfollow_method = make_unfollow_method(model)
    User.add_to_class('unfollow_' + model_name, unfollow_method)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from followit.compat import USER_MODEL_CLASS_NAME

class FollowRecord(models.Model):
    user = models.ForeignKey(USER_MODEL_CLASS_NAME)
    content_type = models.ForeignKey(
                            ContentType,
                            related_name='followed_record_contenttype'
                        )
    object_id = models.PositiveIntegerField(db_index=True)



import random
import string

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'TEST_NAME': ':memory:',
    },
}

MIDDLEWARE_CLASSES = ()

SECRET_KEY = ''.join([random.choice(string.ascii_letters) for x in range(40)])

TEST_RUNNER = 'django.test.runner.DiscoverRunner'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'followit',
    'followit.tests'
)



from django.conf.urls import patterns, url

urlpatterns = patterns('',
    url(
        r'^follow/(?P<model_name>\w+)/(?P<object_id>\d+)/$',
        'followit.views.follow_object',
        name = 'follow_object'
    ),
    url(
        r'^unfollow/(?P<model_name>\w+)/(?P<object_id>\d+)/$',
        'followit.views.unfollow_object',
        name = 'unfollow_object'
    ),
    url(
        r'^toggle-follow/(?P<model_name>\w+)/(?P<object_id>\d+)/$',
        'followit.views.toggle_follow_object',
        name='toggle_follow_object'
    )
)



"""utility functions used by the :mod:`followit`
"""
import functools
import simplejson
from django.http import HttpResponse
import followit

def get_object(model_name, object_id):
    """not a view, just a function, returning an object"""
    model = followit.REGISTRY[model_name]
    return model.objects.get(id = object_id)

def del_related_objects_cache(model):
    try:
        del model._meta._related_objects_cache
    except AttributeError:
        pass
    try:
        del model._meta._related_objects_proxy_cache
    except AttributeError:
        pass

def followit_ajax_view(view_func):
    """decorator that does certain error checks on the input
    and serializes response as json

    in the case of error, json output will contain
    """
    @functools.wraps(view_func)
    def wrapped_view(request, model_name = None, object_id = None):
        try:
            assert(model_name in followit.REGISTRY)
            assert(request.user.is_authenticated())
            assert(request.method == 'POST')
            assert(request.is_ajax())
            data = view_func(request, model_name, object_id)
        except Exception, e:
            data = {'status': 'error', 'error_message': unicode(e)}

        return HttpResponse(simplejson.dumps(data), mimetype = 'application/json')
    return wrapped_view

def post_only(view_func):
    """simple decorator raising assertion error when method is not 'POST"""
    @functools.wraps(view_func)
    def wrapped_view(request, *args, **kwargs):
        assert(request.method == 'POST')
        return view_func(request, *args, **kwargs)
    return wrapped_view
 



from django.apps import AppConfig, apps
from django.contrib.auth import get_user_model

class TestConfig(AppConfig):
    name = 'followit.tests'

    def ready(self):
        import followit
        Car = apps.get_model('tests.Car')
        Alligator = apps.get_model('tests.Alligator')
        followit.register(Car)
        followit.register(Alligator)
        followit.register(get_user_model())#to test following users

default_app_config = 'followit.tests.TestConfig'



"""Views for the ``followit`` app,
all are ajax views and return application/json mimetype
"""
from followit import utils

@utils.followit_ajax_view
@utils.post_only
def follow_object(request, model_name = None, object_id = None):
    """follows an object and returns status:
    * 'success' - if an object was successfully followed
    * the decorator takes care of the error situations
    """
    obj = utils.get_object(model_name, object_id)
    follow_func = getattr(request.user, 'follow_' + model_name)
    follow_func(obj)
    return {'status': 'success'}


@utils.followit_ajax_view
@utils.post_only
def unfollow_object(request, model_name = None, object_id = None):
    """unfollows an object and returns status 'success' or
    'error' - via the decorator :func:`~followit.utils.followit_ajax_view`
    """
    obj = utils.get_object(model_name, object_id)
    unfollow_func = getattr(request.user, 'unfollow_' + model_name)
    unfollow_func(obj)
    return {'status': 'success'}


@utils.followit_ajax_view
@utils.post_only
def toggle_follow_object(request, model_name = None, object_id = None):
    """if object is followed then unfollows
    otherwise follows

    returns json
    {
        'status': 'success', # or 'error'
        'following': True, #or False
    }

    
    unfollows an object and returns status 'success' or
    'error' - via the decorator :func:`~followit.utils.followit_ajax_view`
    """
    obj = utils.get_object(model_name, object_id)
    if request.user.is_following(obj):
        toggle_func = getattr(request.user, 'unfollow_' + model_name)
        following = False
    else:
        toggle_func = getattr(request.user, 'follow_' + model_name)
        following = True

    toggle_func(obj)
    return {
        'status': 'success',
        'following': following
    }



import ez_setup
ez_setup.use_setuptools()
from setuptools import setup, find_packages
import sys

setup(
    name="django-followit",
    version='0.2.0',
    description='A Django application that allows users to follow django model objects',
    packages=find_packages(),
    author='Evgeny.Fadeev',
    author_email='evgeny.fadeev@gmail.com',
    license='BSD License',
    keywords='follow, database, django',
    url='https://github.com/ASKBOT/django-followit',
    include_package_data=True,
    install_requires=['simplejson',],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.4',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: JavaScript',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    ],
    long_description="""The ``followit`` django app allows to easily set up a
capability for the site users to follow various things on the site,
represented by django model ``followit.models.FollowRecord`` 
using the ``django.contrib.models.ContentTypes`` system.

Release Notes
=============
Starting the version ``0.2.0``, django-followit supports django 1.8.

Starting the version ``0.1.0``, `django-followit` supports django 1.7.
Also starting this version, method `register` must be called from your app's
`AppConfig.ready()` method.

Starting the version ``0.0.8`` it is not necessary to run ``syncdb`` for this app,
but instead run the ``migrate`` command.

Setup
========

To the INSTALLED_APPS in your ``settings.py`` add entry ``'followit'``.
Run `python manage.py migrate followit`

Then, in the body of `AppConfig.ready` method, add::
    import followit
    followit.register(Thing)

Not it will be possible for the user to follow instances of ``SomeModel``.

If you decide to allow following another model, just add another
``followit.register(...)`` statement.

Usage
============

Examples below show how to use ``followit`` (assuming that model ``Thing``
is registered with ``followit`` in your ``models.py``::
    bob.follow_thing(x)
    bob.unfollow_thing(x)
    things = bob.get_followed_things()
    x_followers = x.get_followers()

Available urls from the `followit/urls.py`::
    /follow/<model_name>/<item_id>/
    {% url follow_object "somemodel" item_id %} #model name lower case

    /unfollow/<model_name>/<item_id>/
    {% url unfollow_object "somemodel" item_id %} #lower case model name

    /toggle-follow/<model_name>/<item_id>/
    {% url toggle_follow_object "somemodel" item_id %} #lower case model name


``followit`` does not yet provide template tags.
"""
)



from django.db import models
from followit.compat import get_user_model

class Car(models.Model):
    yeah = models.BooleanField(default = True)

class Alligator(models.Model):
    yeah = models.BooleanField(default = True)

